@@ -31,6 +31,7 @@
 const int RAISE = 1;
 const int LOWER = 0;
 
+di_hash_table *seen_items;
 int last_successful_item = -1;
 
 /* Save default priority, to be able to return to it when we have to lower it */
@@ -57,10 +58,20 @@ int package_array_compare (const void *v1, const void *v2) {
 	p2 = *(di_system_package **)v2;
 
 	int r = p1->installer_menu_item - p2->installer_menu_item;
+	//di_log(DI_LOG_LEVEL_DEBUG, "menu compare: %s (%i) vs %s (%i): %i", 
+	//		p1->p.package, p1->installer_menu_item,
+	//		p2->p.package, p2->installer_menu_item, r);
 	if (r) return r;
 	return strcmp(p1->p.package, p2->p.package);
 }
 
+static void seen_items_key_destroy (void *key)
+{
+	di_rstring *s = key;
+	di_free(s->string);
+	di_free(s);
+}
+
 int isdefault(di_system_package *p) {
 	int check;
 
@@ -70,9 +81,19 @@ int isdefault(di_system_package *p) {
 	return false;
 }
 
+/*
+ * Menu items with a number equal to 99999 or 99900 (deprecated) are
+ * intended only to be shown during package selection in anna, but not
+ * in the main menu, so mark them uninstallable.
+ * For other packages, run the isinstallble maintainer script and check
+ * its return code.
+ */
 bool isinstallable(di_system_package *p) {
 	int check;
 
+	if ((p->installer_menu_item == 99999) || (p->installer_menu_item == 99900))
+		return false;
+
 	check = di_system_dpkg_package_control_file_exec(&p->p, "isinstallable", 0, NULL);
 	if (check <= 0)
 		return true;
@@ -128,14 +149,18 @@ get_default_menu_item(di_slist *list)
 			//di_log(DI_LOG_LEVEL_DEBUG, "not menu item; or not installed");
 			continue;
 		}
-		if (p->installer_menu_item < last_successful_item &&
-		    p->installer_menu_item < NEVERDEFAULT) {
+		if (p->installer_menu_item >= NEVERDEFAULT) {
 			//di_log(DI_LOG_LEVEL_DEBUG, "not in range to be default");
 			continue;
 		}
+		if (p->installer_menu_item < last_successful_item &&
+		    di_hash_table_lookup(seen_items, &p->p.key)) {
+			//di_log(DI_LOG_LEVEL_DEBUG, "menu item is before last_successful_item, and is not new");
+			continue;
+		}
 		/* If menutest says this item should be default, make it so */
 		if (!isdefault(p)) {
-			//di_log(DI_LOG_LEVEL_DEBUG, "isdefalt says no");
+			//di_log(DI_LOG_LEVEL_DEBUG, "isdefault says no");
 			continue;
 		}
 		/* If all of the virtual packages provided by a
@@ -262,6 +287,7 @@ di_system_package *show_main_menu(di_packages *packages, di_packages_allocator *
 		di_log(DI_LOG_LEVEL_WARNING, "Internal error! Cannot find \"%s\" in menu.", s);
 	}
 
+	free(s);
 	free(menu);
 	free(package_array);
 
@@ -297,8 +323,10 @@ static int satisfy_virtual(di_system_package *p) {
 			if (dep->p.status != di_package_status_installed) {
 				switch (di_config_package(dep, satisfy_virtual)) {
 					case -1:
+						di_free(menu);
 						return -1;
 					case EXIT_BACKUP:
+						di_free(menu);
 						return EXIT_BACKUP;
 				}
 			}
@@ -363,8 +391,10 @@ static int satisfy_virtual(di_system_package *p) {
 			debconf_capb(debconf, "backup");
 			debconf_subst(debconf, MISSING_PROVIDE, "CHOICES", menu);
 			debconf_input(debconf, priority, MISSING_PROVIDE);
-			if (debconf_go(debconf) != 0)
+			if (debconf_go(debconf) != 0) {
+				di_free(menu);
 				return 0;
+			}
 			debconf_capb(debconf);
 			debconf_get(debconf, MISSING_PROVIDE);
 			s = strdup(debconf->value);
@@ -380,8 +410,12 @@ static int satisfy_virtual(di_system_package *p) {
 				 * the providing packages */
 				switch (di_config_package(dep, satisfy_virtual)) {
 					case -1:
+						di_free(menu);
+						free(s);
 						return -1;
 					case EXIT_BACKUP:
+						di_free(menu);
+						free(s);
 						return EXIT_BACKUP;
 				}
 				if (is_menu_item)
@@ -389,7 +423,7 @@ static int satisfy_virtual(di_system_package *p) {
 			}
 		}
 	}
-	free(menu);
+	di_free(menu);
 	free(s);
 	/* It doesn't make sense to configure virtual packages,
 	 * since they are, well, virtual. */
@@ -499,9 +533,18 @@ static void adjust_default_priority (void) {
 	}
 }
 
+static void restore_default_priority (void) {
+	const char *template = "debconf/priority";
+
+	di_log(DI_LOG_LEVEL_INFO, "Restoring default debconf priority '%s'",
+		debconf_priorities[default_priority] ? debconf_priorities[default_priority] : "(null)");
+	debconf_set(debconf, template, debconf_priorities[default_priority]);
+}
+
 void notify_user_of_failure (di_system_package *p) {
 	char buf[256];
 	
+	set_package_title(p);
 	debconf_capb(debconf);
 	menu_entry(debconf, p, buf, sizeof (buf));
 	debconf_subst(debconf, ITEM_FAILURE, "ITEM", buf);
@@ -568,6 +611,7 @@ int main (int argc __attribute__ ((unused)), char **argv) {
 	di_packages *packages;
 	di_packages_allocator *allocator;
 	int ret;
+	int last_item_backup = 0;
 
 	debconf = debconfclient_new();
 	di_system_init(basename(argv[0]));
@@ -583,9 +627,18 @@ int main (int argc __attribute__ ((unused)), char **argv) {
 
 	menu_startup();
 
+	seen_items = di_hash_table_new_full(di_rstring_hash, di_rstring_equal,
+					    seen_items_key_destroy, NULL);
+
 	allocator = di_system_packages_allocator_alloc ();
 	packages = di_system_packages_status_read_file(DI_SYSTEM_DPKG_STATUSFILE, allocator);
 	while ((p=show_main_menu(packages, allocator))) {
+		di_slist_node *node;
+
+		if (p->installer_menu_item < NEVERDEFAULT && last_item_backup) {
+			restore_default_priority();
+			last_item_backup = 0;
+		}
 		ret = do_menu_item(p);
 		adjust_default_priority();
 		switch (ret) {
@@ -597,19 +650,34 @@ int main (int argc __attribute__ ((unused)), char **argv) {
 					//di_log(DI_LOG_LEVEL_DEBUG, "Installed package '%s', raising last_successful_item to %d", p->p.package, p->installer_menu_item);
 				}
 				else {
+					/* Do not raise the priority after it has been set by cdebconf-priority. */
+					if (0 == strcmp("cdebconf-priority", p->p.package)) {
+						last_item_backup = 0;
+					}
 					// di_log(DI_LOG_LEVEL_DEBUG, "Installed package '%s' but no raise since %d >= %i", p->p.package, p->installer_menu_item, NEVERDEFAULT);
 				}
 				break;
 			case EXIT_BACKUP:
 				di_log(DI_LOG_LEVEL_INFO, "Menu item '%s' succeeded but requested to be left unconfigured.", p->p.package); 
 				modify_debconf_priority(LOWER);
+				last_item_backup = 1;
 				break;
 			default:
 				di_log(DI_LOG_LEVEL_WARNING, "Menu item '%s' failed.", p->p.package);
 				notify_user_of_failure(p);
 				modify_debconf_priority(LOWER);
 		}
-		
+
+		/* Remember all the packages we've seen so far */
+		for (node = packages->list.head; node; node = node->next) {
+			di_system_package *seen = node->data;
+			di_rstring *seen_name = di_new0(di_rstring, 1);
+			seen_name->string = di_stradup(seen->p.key.string,
+						       seen->p.key.size);
+			seen_name->size = seen->p.key.size;
+			di_hash_table_insert(seen_items, seen_name, seen_name);
+		}
+
 		di_packages_free (packages);
 		di_packages_allocator_free (allocator);
 		allocator = di_system_packages_allocator_alloc ();
@@ -622,6 +690,8 @@ int main (int argc __attribute__ ((unused)), char **argv) {
 	return EXIT_FAILURE;
 }
 
+int di_config_package_depth=0;
+
 /*
  * Configure all dependencies, special case for virtual packages.
  * This is done depth-first.
@@ -633,10 +703,10 @@ static int di_config_package(di_system_package *p,
 	di_slist_node *node;
 	di_system_package *dep;
 
-	//di_log(DI_LOG_LEVEL_DEBUG, "configure %s, status: %d\n", p->p.package, p->p.status);
+	//di_log(DI_LOG_LEVEL_DEBUG, "configure %s, status: %d", p->p.package, p->p.status);
 
 	if (p->p.type == di_package_type_virtual_package) {
-		//di_log(DI_LOG_LEVEL_DEBUG, "virtual package %s\n", p->p.package);
+		//di_log(DI_LOG_LEVEL_DEBUG, "virtual package %s", p->p.package);
 		if (virtfunc)
 			return virtfunc(p);
 		else
@@ -653,7 +723,14 @@ static int di_config_package(di_system_package *p,
 		if (d->type != di_package_dependency_type_depends)
 			continue;
 		/* Recursively configure this package */
-		switch (di_config_package(dep, virtfunc)) {
+		di_config_package_depth++;
+		if (di_config_package_depth > 1000) {
+			di_log(DI_LOG_LEVEL_WARNING, "Deep recursion configuring package %s (dep loop?)", p->p.package);
+			return -1;
+		}
+		ret = di_config_package(dep, virtfunc);
+		di_config_package_depth--;
+		switch (ret) {
 			case -1:
 				return -1;
 			case EXIT_BACKUP:
